%% test 4-point homography

clc
clearvars
close all
%%  given points and homography

% points (non-homogeneous)
Xe  = randn(2,4);
Xe  = [1,-1,-1,1;1,1,-1,-1]/2;
% homography
Ht  = eye(3)+randn(3)/2;
%Ht  = [1,2,3;4,5,6;7,8,1];
%Ht=eye(3);
Ht  = Ht / norm(Ht(:))*sign(det(Ht));
% noise standsrd devitation
sigma=0.001;
% CovM (non-homogeneous -> homogeneous)
CC=randn(8);Cxxe=sigma^2*(eye(8)+CC*CC');
Cxx = zeros(12);
for i=1:4
    for j = 1:4
        Cxx(3*i-2:3*i-1,3*j-2:3*j-1)=Cxxe(2*i-1:2*i,2*j-1:2*j);
    end
end

%% generate noisy data, following homogeneous coordinates
Xh  =[Xe;ones(1,4)];
Yht = Ht*Xh;
Yht = Yht./([1;1;1]*Yht(3,:));
yht = Yht(:);
yh  = yht+ rand_gauss(zeros(12,1),Cxx,1);
Yh  = reshape(yh,3,4);
Ye  = Yh(1:2,:)./([1;1]*Yh(3,:));

% plot

figure
dd=0.1;
subplot(1,2,1)
plot(Xe(1,:),Xe(2,:),'.b','markersize',25)
for i=1:4
    text(Xe(1,i)+dd,Xe(2,i)+dd,num2str(i))
end
xlim([min(Xe(1,:))-0.5,max(Xe(1,:))+0.5])
ylim([min(Xe(2,:))-0.5,max(Xe(2,:))+0.5])
axis equal
subplot(1,2,2)
plot(Ye(1,:),Ye(2,:),'.b','markersize',25)
for i=1:4
    text(Ye(1,i)+dd,Ye(2,i)+dd,num2str(i))
end
xlim([min(Ye(1,:))-0.5,max(Ye(1,:))+0.5])
ylim([min(Ye(2,:))-0.5,max(Ye(2,:))+0.5])
axis equal

%% estimate homography

disp('----- Homography using homogeneous coordinates of four points -----')

[He,Chh]=calc_homography_4_points_homogeneous(Xh,Yh,Cxx,Ht);


Hh=He;
Ch = Chh;

%% perform evaluation
% condition number of H
cond_H=cond(He)
% sqrt of condition number of Chh, ratio of standard deviations
eChh=sort(eig(Chh));
sqrt_cond_Chh=round(sqrt(eChh(9)/eChh(2)))

% Xhi-2-test: estimated vs. true homography
checkHt=Ht/norm(Ht(:))*sign(det(Ht));
checkHe=He/norm(He(:))*sign(det(He));
dH = (checkHt-checkHe);
X = dH(:)'*pinv(Chh)*dH(:);
disp(['test un-conditioned X = ', num2str(X),...
    ' interval = [',num2str(chi2inv(0.005,8)),',',num2str(chi2inv(0.995,8)),']'])


%% same for non-homogeneous coordinates

disp('----- Homography using non-homogeneous coordinates of four points -----')

[He,Chh]=calc_homography_4_points(Xe,Ye,Cxxe,Ht);

% condition number of H
cond_H=cond(He)
% sqrt of condition number of Chh, ratio of standard deviations
eChh=sort(eig(Chh));
sqrt_cond_Chh=round(sqrt(eChh(9)/eChh(2)))

% Xhi-2-test: estimated vs. true homography
checkHt=Ht/norm(Ht(:))*sign(det(Ht));
checkHe=He/norm(He(:))*sign(det(He));
dH = (checkHt-checkHe);
X = dH(:)'*pinv(Chh)*dH(:);
disp(['test un-conditioned X = ', num2str(X),...
    ' interval = [',num2str(chi2inv(0.005,8)),',',num2str(chi2inv(0.995,8)),']'])

%% check identity

disp('---------- compare results ----------')
dH=Hh-He
dC = Ch-Chh